# 矩阵形式的感知机

## 矩阵形式的感知机模型公式表和符号表

### 公式表
|模型名称|模型分类决策函数|模型损失函数|模型损失函数的梯度|
|-|-|-|-|
|感知机|$\hat{Y}=where(XW \geq 0, +1, -1)$|$Loss =\frac{1}{m} sum(max(0, -Y\odot (XW)))$|$\dfrac{\partial Loss}{\partial W} = - \frac{1}{m}sum(Y_{p}\odot X_{p}, axis=0)$|

### 符号表
|符号|向量形式|维度|意义
|-|-|-|-|
|$x$|$[x_0,x_1,...,x_n]^T$|(n, 1)|$x$ 表示一个样本实例， $x_i$ 表示向量的第 $i$ 维特征。|
|$X$|$[{(x^{(1)})}^T,{(x^{(2)})}^T,...{(x^{(m)})}^T]$| (m,n)|样本矩阵，一共 $m$ 个样例，每个样例有 $n$ 个特征值，$x^{(i)}$ 表示第 $i$ 个样例。|
|$w$|${[w_{0},w_{1},...,w_{n}]}^T$| (n,1)|$w$ 表示 $x$ 的权重系数。|
|$W$|${[w_{0},w_{1},...,w_{n}]}^T$|(n,1)|权重矩阵，与 $w$ 意义一致。|
|$y$|标量|(1,1)|样例的类别，1 为正类， -1 负类|
|$\hat y$|标量|(1,1)|预测样例的类别，1 为正类， -1 负类|
|$Y$|$[y^{(1)},y^{(2)},...,y^{(m)}]$|(m,1)|标签矩阵|
|$\hat{Y}$|$[\hat{y}^{(1)},\hat{y}^{(2)},...,\hat{y}^{(m)}]$|(m,1)|预测矩阵，m 个样例对应的类别|

符号定义规则：一个单独的向量比如 $x$ 一般是列向量，但是把单独的一个向量放进矩阵中时，要把向量先转置。


## 感知器模型

感知器（Perceptron）由Frank Roseblatt 于1957年提出，是一种广泛使用的线性分类器。感知器可谓是最简单的人工神经网络，只有一个神经元。感知器是对生物神经元的简单数学模拟，有与生物神经元相对应的部件，如权重（突触）、偏置（阈值）及激活函数（细胞体），输出为0或1。感知器是一种简单的两类线性分类模型。


## 感知器的决策函数

一般形式的感知器决策函数：
$$
\hat{y}^{(i)} =
\begin{cases}
  -1 & \text{if}\quad  w^{T}x^{(i)} < 0, \\
  +1 & \text{if}\quad  w^{T}x^{(i)} \geq 0.
\end{cases}
$$

矩阵形式的感知器决策函数
$$
\hat{Y}=where(XW \geq 0, +1, -1)
$$

## 感知器的损失函数

一般形式的感知器的损失函数，

$$
Loss =\frac{1}{m} \sum_{i=1}^m max(0, -y^{(i)}(w^Tx^{(i)}))
$$

矩阵形式的感知器的损失函数，

$$
Loss =\frac{1}{m} sum(max(0, -Y\odot (XW)))
$$

## 感知器的损失函数的梯度向量

一般形式的感知机的损失函数的梯度向量：

$$
\dfrac{\partial Loss}{\partial w} =\frac{1}{m} \sum_{i=1}^m
\begin{cases}
  0 & \text{if}\quad  w^{T}x^{(i)} > 0,, \\ - y^{(i)} * x^{(i)}  & \text{if}\quad  w^{T}x^{(i)} \le 0,.
\end{cases}
$$

矩阵形式的感知器的损失函数的梯度向量：

首先定义不满足超平面条件的样例和对应标签构成的矩阵 $X_{p}$ 和 $Y_{p}$：

$$X_{p}:\quad  x^{(i)}\in X_{p} \quad \text{if} \quad w^{T}x^{(i)} \le 0$$

$$Y_{p}:\quad y^{(i)}\in Y_{p} \quad \text{if} \quad w^{T}x^{(i)} \le 0$$


$$\dfrac{\partial Loss}{\partial W} = - \frac{1}{m}sum(Y_{p}\odot X_{p}, axis=0) $$
